# Repository path:    $HeadURL: https://90.145.156.222/DMS-Houtribdijk/code/python-ftp-svn/source/readRaws/ReadRaws.py $
# Last committed:     $Revision: 417449 $; $Author: vvuik $; $Date: 2019-11-12 15:00:41 +0100 (di, 12 nov 2019) $
#
import csv
import numpy as np
import pandas as pd
import re
from operator import itemgetter
from itertools import chain
from os import path, makedirs, remove
from shutil import copyfile
from re import sub
from misc.REMatcher import REMatcher


def copy_file_to_failed_folder(source, settings):
    target = sub(settings["import_folder"], settings["import_failed"], source)
    try:
        makedirs(path.dirname(target))
    except FileExistsError:
        pass
    try:
        copyfile(source, target)
    except FileExistsError:
        remove(target)
        copyfile(source, target)


class ReadRawStb:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawStb.init: $Revision: 417449 $")

        # Column names and parameter mapping: replace dots in de Delft-FEWS parameters with underscores
        self._par_map = {'EtroLevel': 'EtroLevel'}

        col_names = ['TIMESTAMP', 'RECORD'] + list(self._par_map)

        try:
            # Create a data frame of the data file (use the column names and skip the first 4 rows)
            self._df = pd.read_csv(fn, names=col_names, header=None, skiprows=4, na_values=['NAN'])
        except FileNotFoundError:
            logger.error("ReadRawStb.init: File {} not found".format(fn))
            self._data_read_successful = False
        except pd.errors.ParserError:
            logger.error("ReadRawStb.init: Error reading file {}".format(fn))
            self._data_read_successful = False
        else:
            try:
                # Subtract the year, month, day, time from the column 'TIME' and store in columns as integer
                self._df['YEAR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(0, 4))
                self._df['MONTH'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(5, 7))
                self._df['DAY'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(8, 10))
                self._df['HOUR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(11, 13))
                self._df['MINUTE'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(14, 16))
                self._df['SECOND'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(17, 22))

                # Create a column with the datetime
                self._df['DATETIME'] = pd.to_datetime(self._df.loc[:, ['YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND']])
                # Use the datetime as the index of the data frame
                self._df.index = self._df.pop('DATETIME')
                # Delete the columns 'TIME', 'YEAR', 'MONTH', etc
                self._df.drop(['TIMESTAMP', 'YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND', 'RECORD'], axis=1, inplace=True)
                self._df.sort_index(inplace=True)

                # cm naar m
                self._df['EtroLevel'] = self._df['EtroLevel'].astype(float)/100
                self._df.fillna(-9999, inplace=True)
                self._data_read_successful = True
                
            except AttributeError:
                logger.error("ReadRawStb.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    def get_data(self):
        return self._df

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        return self._par_map

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 1


class ReadRawMpp:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawMpp.init: $Revision: 417449 $")

        # Column names and parameter mapping: replace dots in de Delft-FEWS parameters with underscores
        self._par_map = {'YSI_Temp': 'T_0', 'YSI_Cond': 'C_0', 'YSI_Ph': 'Ph_0', 'YSI_Turb': 'OPAQ_0', 'YSI_Chl': 'Chl_0'}

        try:
            # Create a data frame of the data file (use the column names and skip the first 4 rows)
            self._df = pd.read_csv(fn, header=None, skiprows=4, na_values=['NAN', '+INF', '-INF'])
            # Check whether the oxygen columns are present in the file or not
            if len(self._df.columns) > 8:
                self._par_map['YSI_DOr'] = 'O2_0_perc'
                self._par_map['YSI_DOmg'] = 'O2_0_mgl'

            col_names = ['TIMESTAMP', 'RECORD'] + list(self._par_map) + ['YSI_Samples']
            self._df.columns =  col_names
        except FileNotFoundError:
            logger.error("ReadRawMpp.init: File {} not found".format(fn))
            self._data_read_successful = False
        else:
            try:
                # Subtract the year, month, day, time from the column 'TIME' and store in columns as integer
                self._df['YEAR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(0, 4))
                self._df['MONTH'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(5, 7))
                self._df['DAY'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(8, 10))
                self._df['HOUR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(11, 13))
                self._df['MINUTE'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(14, 16))
                self._df['SECOND'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(17, 19))

                self._df['DATETIME'] = pd.to_datetime(self._df.loc[:, ['YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND']])
                # Use the datetime as the index of the data frame
                self._df.index = self._df.pop('DATETIME')
                # Delete the columns 'TIME', 'YEAR', 'MONTH', etc
                self._df.drop(['TIMESTAMP', 'YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND', 'RECORD'], axis=1, inplace=True)

                self._df.fillna(-9999, inplace=True)
                self._data_read_successful = True
                
            except AttributeError:
                logger.error("ReadRawMpp.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    def get_data(self):
        return self._df

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        return self._par_map

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 1


class ReadRawAlt:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawAlt.init: $Revision: 417449 $")

        # Column names and parameter mapping: replace dots in de Delft-FEWS parameters with underscores
        self._par_map = {'ISA_DIST': 'ALTdist'}

        col_names = ['TIME', 'RECORD', 'ISA_DIST', 'ISA_Pitch','ISA_Roll']

        try:
            # Create a data frame of the data file (use the column names and skip the first 4 rows)
            self._df = pd.read_csv(fn, names=col_names, header=None, skiprows=4)

        except FileNotFoundError:
            logger.error("ReadRawAlt.init: File {} not found".format(fn))
        else:
            try:
                # Subtract the year, month, day, time from the column 'TIME' and store in columns as integer
                self._df['YEAR'] = pd.to_numeric(self._df['TIME'].str.slice(0, 4))
                self._df['MONTH'] = pd.to_numeric(self._df['TIME'].str.slice(5, 7))
                self._df['DAY'] = pd.to_numeric(self._df['TIME'].str.slice(8, 10))
                self._df['HOUR'] = pd.to_numeric(self._df['TIME'].str.slice(11, 13))
                self._df['MINUTE'] = pd.to_numeric(self._df['TIME'].str.slice(14, 16))
                self._df['SECOND'] = pd.to_numeric(self._df['TIME'].str.slice(17, 19))

                # Create a column with the datetime
                self._df['DATETIME'] = pd.to_datetime(self._df.loc[:, ['YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND']])
                # Use the datetime as the index of the data frame
                self._df.index = self._df.pop('DATETIME')
                # Delete the columns 'TIME', 'YEAR', 'MONTH', etc
                self._df.drop(['TIME', 'YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND'], axis=1, inplace=True)
                self._df.sort_index(inplace=True)

                self._df['ISA_DIST'] = self._df['ISA_DIST'].astype(float)
                
                self._df.fillna(-9999, inplace=True)
                self._data_read_successful = True
                
            except AttributeError:
                logger.error("ReadRawAlt.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    def get_data(self):
        return self._df

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        return self._par_map

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 1


class ReadRawAdv:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawAdv.init: $Revision: 417449 $")
        
        # Column names and parameter mapping: replace dots in de Delft-FEWS parameters with underscores
        self._par_map = {'Pressure': 'D_0_water',
                         'U_east': 'S_0_east', 'U_north': 'S_0_north', 'U_up': 'S_0_up',
                         'Amp1': 'Amp1_0', 'Amp2': 'Amp2_0', 'Amp3': 'Amp3_0',
                         'Corr1': 'Corr1_0', 'Corr2': 'Corr2_0', 'Corr3': 'Corr3_0'}
        self._par_map_system = {'BatteryVoltage': 'V_0_battery', 'Soundspeed': 'S_0_sound',
                              'Heading': 'Heading_0', 'Pitch': 'Pitch_0', 'Roll': 'Roll_0',
                              'Temperature': 'T_0', 'ErrorCode': 'ErrCode_0', 'Status': 'Status_0'}

        try:
            # Read csv
            lines = []
            with open(fn, 'r') as csvfile:
                data = csv.reader(csvfile, delimiter=',')
                for line in data:
                    lines.append(line)
        except FileNotFoundError:
            logger.error("ReadRawAdv.init: File {} not found".format(fn))
            self._data_read_successful = False
            
        else:
            try:
        
                # system data
                index_system = [idx for idx, line in enumerate(lines) if line[-1] == 'Checksum ']
                if (index_system[1] - index_system[0]) == 2:
                    del index_system[0]
                indices = np.arange(3,17)
                get_items = itemgetter(*indices)
                data = [get_items(lines[i+1]) for i in index_system]

                self._df_system = pd.DataFrame(
                        data = data,
                        columns = ['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second', 'BatteryVoltage', 'Soundspeed', 
                                   'Heading', 'Pitch', 'Roll', 'Temperature', 'ErrorCode', 'Status']).astype('float')
    
                # given with 0.1 precision
                self._df_system[['BatteryVoltage', 'Soundspeed', 'Heading', 'Pitch', 'Roll']] /= 10
                # given with 0.01 precision
                self._df_system[['Temperature']] /= 100
                # errorcode and status
                self._df_system[['ErrorCode', 'Status']] += 9E8

                # make datetime index
                self._df_system['Year'] += 2000
                self._df_system['Datetime'] = pd.to_datetime(self._df_system.loc[:,['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second']])
                self._df_system.index = self._df_system.pop('Datetime')
                self._df_system.drop(['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second'], axis=1, inplace=True)

                # measurement data 
                index = [idx for idx, line in enumerate(lines) if line[-1] == 'pressureLSW']
                indices = [4, 5]
                get_items = itemgetter(*indices)
                indices_v = [0, 1, 2]
                get_items_v = itemgetter(*indices_v)
                data = [get_items(lines[i+1]) + get_items_v(lines[i+3]) + get_items_v(lines[i+5]) + get_items_v(lines[i+7]) for i in index]

                self._df = pd.DataFrame(
                        data = data,
                        columns = ['PressureMSB', 'PressureLSW', 'Vel1', 'Vel2', 'Vel3', 
                                   'Amp1', 'Amp2', 'Amp3', 'Corr1', 'Corr2', 'Corr3']).astype('float')

                # given with 0.001 precision
                self._df[['Vel1', 'Vel2', 'Vel3']] /= 1000
                
                # Given in System Integrator Manual: Pressure = 65536 × PressureMSB + PressureLSW (pay attention that the units have changed)       
                # Convert to hPa
                self._df['Pressure'] = (65536 * self._df['PressureMSB'] + self._df['PressureLSW'])/10
                self._df.drop(['PressureMSB', 'PressureLSW'], axis=1, inplace=True)

                # Add index (datetimes to df)
                start = sum(i < index_system[0] for i in index)
                if start >= 1:
                    time_start = min(self._df_system.index) - pd.to_timedelta((start-1)*250, unit='ms')
                else:
                    time_start = min(self._df_system.index) + pd.to_timedelta(250, unit='ms')
                    
                end = sum(i > index_system[-1] for i in index)
                time_end = max(self._df_system.index) + pd.to_timedelta(end*250, unit='ms')
                
                datetime_index = pd.date_range(start=time_start, end=time_end, freq='250ms')
                self._df.index = datetime_index

                # transformation of velocity
                
                self._df[['Heading', 'Pitch', 'Roll']] = self._df_system[['Heading', 'Pitch', 'Roll']]
                # interpolate the heading, pitch, and roll
                self._df.interpolate(method='index', axis=0, inplace=True, limit_area='inside')
                # fill other values in dataframe that are outside the range of the times of the system data
                self._df.fillna(method = 'ffill', inplace=True)
                self._df.fillna(method = 'bfill', inplace=True)

                velocities = []
                for row in self._df.itertuples():

                    hh = np.pi*(row.Heading-90)/180
                    pp = np.pi*row.Pitch/180
                    rr = np.pi*row.Roll/180
                
                    # cos and sin of angles
                    c_hh, s_hh = np.cos(hh), np.sin(hh)
                    c_pp, s_pp = np.cos(pp), np.sin(pp)
                    c_rr, s_rr = np.cos(rr), np.sin(rr)  
                    
                    x = row.Vel1
                    y = row.Vel2
                    z = row.Vel3
    
                    # Heading matrix
                    H = np.array([[c_hh,             s_hh,       0],
                                  [-s_hh,            c_hh,       0],
                                  [0,                0,          1]])
                    # Tilt matrix
                    P = np.array([[c_pp,             -s_pp*s_rr,   -c_rr*s_pp],
                                  [0,                c_rr,         -s_rr],
                                  [s_pp,             s_rr*c_pp,    c_pp*c_rr]])
                    # No transformation for beam to xyz (already in xyz)
                    T = np.array([[1,                0,                   0],
                                  [0,                1,                   0],
                                  [0,                0,                   1]])
                    R = np.matmul(np.matmul(H, P), T)
    
                    velocities.append(np.matmul(R, np.array([x, y, z])))    
    
                U = np.vstack(velocities)
                self._df = self._df.assign(U_east = U[:,0], U_north = U[:,1], U_up = U[:,2])
                self._df.drop(['Vel1', 'Vel2', 'Vel3', 'Heading', 'Pitch', 'Roll'], axis=1, inplace=True)

                self._df.fillna(value=-999, inplace=True)
                self._data_read_successful = True

            except (AttributeError, ValueError, SystemError, IndexError):
                logger.error("ReadRawAdv.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    # high frequent
    def get_data(self):
        return self._df

    def get_data2(self):
        return self._df_system

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        return self._par_map

    def get_par_map2(self):
        return self._par_map_system

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 2


class ReadRawAqd:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawAqd.init: $Revision: 417449 $")

        self._par_map = {}

        try:
            lines = []
            with open(fn, 'r') as csvfile:
                data = csv.reader(csvfile, delimiter=',')
                for line in data:
                    lines.append(line)

        except FileNotFoundError:
            logger.error("ReadRawAqd.init: File {} not found".format(fn))
            self._data_read_successful = False
        else:
            try:
                indices = [3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 16, 18, 19]   
                get_items = itemgetter(*indices)
                data = [get_items(lines[i]) for i in np.arange(3, len(lines), 13)]
                self._df = pd.DataFrame(
                        data = data,
                        columns = ['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second', 'Millisecond', 
                                   'V_0_battery', 'S_0_sound', 'Heading_0', 'Pitch_0', 'Roll_0', 
                                   'PressureMSB', 'PressureLSW', 'T_0']).astype(float)
    
                # are given with 0.1 precision
                self._df[['V_0_battery', 'S_0_sound', 'Heading_0', 'Pitch_0', 'Roll_0']] /= 10    
                # are given with 0.01 precision
                self._df['T_0'] /= 100 

                # Given in System Integrator Manual: Pressure = 65536 × PressureMSB + PressureLSW 
                # Convert to hPa     
                self._df['D_0_water'] = (65536 * self._df['PressureMSB'] + self._df['PressureLSW'])/10
                self._df.drop(['PressureMSB', 'PressureLSW'], axis=1, inplace=True)
                
                # Make list with names of parameters with indication of level
                # Assumes that number of levels can be determined from 5th line of the data
                nr_levels = len(lines[5]) 
                levels_range = np.arange(1, nr_levels)
                levels = [np.str(i) for i in levels_range]

                column_names = {
                    'Sx': ['S' + s + '_0_x' for s in levels],
                    'Sy': ['S' + s + '_0_y' for s in levels],
                    'Sz': ['S' + s + '_0_z' for s in levels],
                    'Amp1': ['Amp1' + s + '_0' for s in levels],
                    'Amp2': ['Amp2' + s + '_0' for s in levels],
                    'Amp3': ['Amp3' + s + '_0' for s in levels],
                    'Corr1': ['Corr1' + s + '_0' for s in levels],
                    'Corr2': ['Corr2' + s + '_0' for s in levels],
                    'Corr3': ['Corr3' + s + '_0' for s in levels],
                    'Seast': ['S' + s + '_0_east' for s in levels],
                    'Snorth': ['S' + s + '_0_north' for s in levels],
                    'Sup': ['S' + s + '_0_up' for s in levels]
                    }
                
                # create dictionary
                variables = ['Sx', 'Sy', 'Sz', 'Amp1', 'Amp2', 'Amp3', 'Corr1', 'Corr2', 'Corr3']
                crsvars = ['Seast', 'Snorth', 'Sup']
                data = {var: [] for var in variables}

                # add columns to df
                namenlijst = []
                for var in variables + crsvars:
                    namenlijst.extend(column_names[var])
                self._df = self._df.reindex(columns = self._df.columns.tolist() + namenlijst)

                # Read measurement data
                indices = [0, 1, 2, 4, 5, 6, 8, 9, 10]
                try:
                    for i in np.arange(5, len(lines), 13):
                        for j, var in zip(indices, variables):
                            data[var].append(lines[i + j][:-1])
                except IndexError:
                    logger.warning("ReadRawAqd.init: File ({}) not complete".format(fn))
                    copy_file_to_failed_folder(fn, settings)
                finally:
                    # Add measurement data to df
                    for var in variables:
                        self._df.loc[:, column_names[var]] = np.array(data[var], dtype=float)

                    # convert velocities from mm/s to m/s
                    self._df[namenlijst[:45]] /= 1000

                    # Make datetime index
                    self._df['Year'] = self._df['Year'] + 2000
                    self._df['Datetime'] = pd.to_datetime(self._df.loc[:,['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second', 'Millisecond']])
                    self._df.index = self._df.pop('Datetime')
                    self._df.drop(['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second', 'Millisecond'], axis=1, inplace=True)
                    del self._df.index.name

                    ### Transformation of coordinate system ###

                    # create dictionary with key for each datetime
                    velocities = {idx: [] for idx in self._df.index}

                    for row in self._df.itertuples():

                        # heading, pitch, roll
                        hh = np.pi*(row.Heading_0-90)/180
                        pp = np.pi*row.Pitch_0/180
                        rr = np.pi*row.Roll_0/180

                        # cos and sin of angles
                        c_hh, s_hh = np.cos(hh), np.sin(hh)
                        c_pp, s_pp = np.cos(pp), np.sin(pp)
                        c_rr, s_rr = np.cos(rr), np.sin(rr)

                        # Heading matrix
                        H = np.array([[c_hh,             s_hh,       0],
                                      [-s_hh,            c_hh,       0],
                                      [0,                   0,       1]])
                        # Tilt matrix
                        P = np.array([[c_pp,             -s_pp*s_rr,    -c_rr*s_pp],
                                      [0,                      c_rr,         -s_rr],
                                      [s_pp,              s_rr*c_pp,     c_pp*c_rr]])
                        # No transformation for beam to xyz (already in xyz)
                        T = np.array([[1,                0,                   0],
                                      [0,                1,                   0],
                                      [0,                0,                   1]])
                        R = np.matmul(np.matmul(H, P), T)

                        for level in levels:
                            Sx = getattr(row, 'S'+level+'_0_x')
                            Sy = getattr(row, 'S'+level+'_0_y')
                            Sz = getattr(row, 'S'+level+'_0_z')

                            velocities[row.Index].append(np.matmul(R, np.array([Sx, Sy, Sz])))

                    arr = [list(chain(*zip(*velocities[idx]))) for idx in self._df.index]
                    self._df[column_names['Seast'] + column_names['Snorth'] + column_names['Sup']] = arr
                    self._df.drop(namenlijst[:45], axis=1, inplace=True)

                    self._data_read_successful = True
            except (AttributeError, ValueError, IndexError):
                logger.error("ReadRawAqd.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    def get_data(self):
        return self._df

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        # return all columns
        par_map = {}
        for col in list(self._df):
            if re.search('^Amp\d', col):
                m = re.match('^Amp(\d)(\d*)_0', col)
                par = "Amp{}_0_n".format(m.group(1)) + "{0:0>2}".format(m.group(2))
                par_map[col] = par
            elif re.search('^Corr\d', col):
                m = re.match('^Corr(\d)(\d*)_0', col)
                par = "Corr{}_0_n".format(m.group(1)) + "{0:0>2}".format(m.group(2))
                par_map[col] = par
            elif re.search('S\d', col):
                m = re.match('^S(\d*)_0_(\w*)', col)
                par = "S_0_{}_n".format(m.group(2)) + "{0:0>2}".format(m.group(1))
                par_map[col] = par
            else:
                par_map[col] = col

        return par_map

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 1                   


class ReadRawAdc:

    def __init__(self, fn, logger, settings):
        self._fn = fn
        logger.debug("ReadRawAdc.init: $Revision: 417449 $")
        
        try:
            if len(pd.read_csv(fn))==0:
                logger.error("ReadRawAdc.init: File ({}) is empty".format(fn))
                self._data_read_successful = False 
            else:
                # Create a data frame of the data file (use the column names and skip the first 4 rows)
                self._df = pd.read_csv(fn, header=0, skiprows=[0,2,3])
        except FileNotFoundError:
            logger.error("ReadRawAdc.init: File {} not found".format(fn))
            self._data_read_successful = False
        except pd.errors.ParserError:
            logger.error("ReadRawAdc.init: File format ({}) does not comply with the definition of the data frame".format(fn))
            self._data_read_successful = False
        else:
            try:
                # Extract the year, month, day, hour, minute, second, from the column 'TIMESTAMP' and store in columns as integer
                self._df['YEAR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(0, 4))
                self._df['MONTH'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(5, 7))
                self._df['DAY'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(8, 10))
                self._df['HOUR'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(11, 13))
                self._df['MINUTE'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(14, 16))
                self._df['SECOND'] = pd.to_numeric(self._df['TIMESTAMP'].str.slice(17, 19))

                self._df['DATETIME'] = pd.to_datetime(self._df.loc[:, ['YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND']])
                # Use the datetime as the index of the data frame
                self._df.index = self._df.pop('DATETIME')
                # Delete the columns 'TIME', 'YEAR', 'MONTH', etc
                self._df.drop(['TIMESTAMP', 'RECORD', 'W_C_Date', 'W_C_Time', 'YEAR', 'MONTH', 'DAY', 'HOUR', 'MINUTE', 'SECOND', 'RECORD'], axis=1, inplace=True)

                # Rename columns to match names in FEWS
                names_Dir = ['Dir_0_n' + '%02d' % (x) for x in range(40)]
                names_Corr = ['Corr_0_n' + '%02d' % (x) for x in range(40)]
                names_S_east = ['S_0_east_n' + '%02d' % (x) for x in range(40)]
                names_S_north = ['S_0_north_n' + '%02d' % (x) for x in range(40)]
                names_S_up = ['S_0_up_n' + '%02d' % (x) for x in range(40)]
                names_S_err = ['S_0_err_n' + '%02d' % (x) for x in range(40)]
                names_Echo1 = ['E1_0_n' + '%02d' % (x) for x in range(40)]
                names_Echo2 = ['E2_0_n' + '%02d' % (x) for x in range(40)]
                names_Echo3 = ['E3_0_n' + '%02d' % (x) for x in range(40)]
                names_Echo4 = ['E4_0_n' + '%02d' % (x) for x in range(40)]

                self._names = ['Heading_0', 'Pitch_0', 'Roll_0', 'T_0', 'S_0_sound', 'status_Bit'] + names_Dir + names_Corr + names_S_east + names_S_north + names_S_up + names_S_err + names_Echo1 + names_Echo2 + names_Echo3 + names_Echo4

                # dictionary of column names and names in FEWS
                columnnames = dict(zip(self._df.columns, self._names))

                self._df.rename(columns=columnnames, inplace=True)

                # Change velocity from mm/s to m/s
                self._df.loc[:, 'S_0_east_n00':'S_0_err_n39'] = self._df.loc[:, 'S_0_east_n00':'S_0_err_n39'] / 1000
                self._df.fillna(value=-999, inplace=True)
                self._df.replace('NAN', '-999', inplace=True)
                self._data_read_successful = True
            except (AttributeError, KeyError, TypeError, NameError):
                logger.error("ReadRawAdc.init: File format ({}) does not comply with the definition of the data frame".format(fn))
                self._data_read_successful = False

    def get_data(self):
        return self._df

    def get_data_read_successful(self):
        return self._data_read_successful

    def get_par_map(self):
        # return all columns except for the levels that are not used
        par_map = {}
        for col in self._names:
            m = REMatcher(col)
            if m.search('n(\d{2})'):
                level = int(m.group(1))
                if level > 23:
                    continue
            par_map[col] = col
        return par_map

    def get_fn(self):
        return self._fn

    def get_number_of_dfs(self):
        return 1
